#!/usr/bin/perl -w

$VERSION = 0.3;

use strict;
use Getopt::Std;
use Carp;
use Digest::MD5;
use vars qw($MD5 $DEBUG $MAIL);  # setup globals under strict

BEGIN
{
    $ENV{'PATH'} = '/sbin:/bin';  # clean out our path
    $ENV{'BASH_ENV'} = '';  # this funkiness fixes weirdness with the subshell
    $ENV{'ENV'} = '';
}

# setup file scoped vars under use strict
my($DB_FILE, $LIST_FILE, $USER);

######################################################################
# info to customize
$USER = '';  # (optional) address to email warnings, try 'root@localhost'

######################################################################
# PATHs, these should be adjusted to match your system
$DB_FILE = "$ENV{'HOME'}/claymore-$::VERSION/light.db";
$LIST_FILE = "$ENV{'HOME'}/claymore-$::VERSION/light.list";
$MAIL = '/bin/mail';

######################################################################
# other globals, you shouldn't have to mess with anything below here
######################################################################
$DEBUG = 0;  # default is off

# initialize the global Digest::MD5 object, used by &checksum
$MD5 = &Digest::MD5::new('Digest::MD5');

# get command line options
my $cmd_line = join(' ', @ARGV);
unless(getopts('d:D:e:hl:mrv')) {
    &usage();
    die "Incorrect option in '$0 $cmd_line'\n";
}
# bomb on help first
if ($Getopt::Std::opt_h) {
    $Getopt::Std::opt_h = 1;  # bogus statement to remove warning
    &usage();
    exit(0);
}
# and version
if ($Getopt::Std::opt_v) {
    $Getopt::Std::opt_v = 1;  # bogus statement to remove warning
    &version();
    exit(0);
}
# get the flags with input
if ($Getopt::Std::opt_D) {
    unless ($Getopt::Std::opt_D =~ /^\d+$/) {
	die "Debug flag ($Getopt::Std::opt_D) not numeric.\n";
    }
    $DEBUG = $Getopt::Std::opt_D;
}
if ($Getopt::Std::opt_d) {
    $DB_FILE = $Getopt::Std::opt_d;
}
if ($Getopt::Std::opt_l) {
    $LIST_FILE = $Getopt::Std::opt_l;
}
if ($Getopt::Std::opt_e) {
    if ($Getopt::Std::opt_e =~ /.*@.*@.*/) {
	die "Option '$Getopt::Std::opt_e' is tainted: only one @ allowed.\n";
    }
    unless ($Getopt::Std::opt_e =~ /.*@.*/) {
	$Getopt::Std::opt_e .= '@localhost';
    }
    if ($Getopt::Std::opt_e =~ /^([\w\.]+\@[\w\.]+)$/) {
	$USER = $1;
    }
    else {
	die "Option '$Getopt::Std::opt_e' is tainted: only a-zA-Z0-9 and " .
	    "dots allowed.\n";
    }
}

# check the input
$LIST_FILE = &un_taint($LIST_FILE);
$DB_FILE = &un_taint($DB_FILE);

# get the flags which modify program behavior
if ($Getopt::Std::opt_m) {
    $Getopt::Std::opt_m = 1;  # bogus statement to remove warning
    &make_list();
    exit(0);
}
if ($Getopt::Std::opt_r) {
    $Getopt::Std::opt_r = 1;  # bogus statement to remove warning
    my @files = &read_list($LIST_FILE);
    &build_db($DB_FILE, @files);
    exit(0);
}

# report on options
if ($DEBUG >= 1) {
    print "LIST_FILE: $LIST_FILE\n";
    print "  DB_FILE: $DB_FILE\n";
}

# run the program
my @files = &read_list($LIST_FILE);
my %db = &read_db($DB_FILE);
my $msg = &test_files(\@files, \%db);
if ($msg and $USER) {
    &send_email($USER, $msg);
}
exit(0);  # we are done


######################################################################
# INTERNAL SUBROUTINES
######################################################################

# this just prints out the usage message
sub usage
{
    print STDERR "Usage: $0 [options], where [options] consist of:\n";
    print STDERR "\t-d filename\tUse filename for the database.\n";
    print STDERR "\t-D lvl\t\tTurn on numeric lvl of debugging.\n";
    print STDERR "\t-e email\tMail the errors to email address too.\n";
    print STDERR "\t-h\t\tPrint this summary of command line options.\n";
    print STDERR "\t-l filename\tUse filename for list of files to check.\n";
    print STDERR "\t-m\t\tPrint a recommended list of files for checking.\n";
    print STDERR "\t-r\t\tRebuild the database.\n";
    print STDERR "\n";
    return;
}

# this just prints out the version
sub version
{
    print STDERR "Claymore ($0) v$::VERSION by Sam Carter\n";
    return;
}

# this just prints out a recommended list of files:
# (/bin /sbin /usr/bin /usr/sbin)
sub make_list
{
    my @dirs = ('/bin', '/sbin', '/usr/bin', '/usr/sbin');
    my $dir;
    foreach $dir (@dirs) {
	opendir(DIR, $dir) or die "Can't open $dir for read: $!\n";
	my @files = readdir(DIR);
	my $file;
	foreach $file (@files) {
	    # only take regular files that match the tainting criteria
	    if (-f "$dir/$file" and "$dir/$file" =~ /^(\/[\w\/\.\-\+]+)$/) {
		print "$dir/$file\n";
	    }
	}
	closedir DIR or die "Can't close $dir from read: $!\n";
    }
    return;
}

# this subroutine reads in a list of files from a flat text list, with one file
# per line, it expects one parameter (the list file), and returns a (possibly
# empty) list
sub read_list
{
    my $FILE = shift;
    my($line, @files);
    open(LIST, "<$FILE") or die "Can't open $FILE for read: $!\n";
    while (defined($line = <LIST>)) {
	next if $line =~ /^\s*$/;
	$line =~ s/\s*#.*$//;
	chomp($line);
	if (-f $line and -r $line) {
	    print "Found file '$line'\n" if $DEBUG >= 2;
	    push(@files, $line);
	}
	else {
	    print "Skipping line '$line'\n" if $DEBUG >= 1;
	}
    }
    close(LIST) or die "Can't close $FILE from read: $!\n";
    return @files;
}

# this builds a hash of filename => md5sum pairs from the file passed in
# as an argument
sub read_db
{
    my $FILE = shift;
    my %db = ();
    open(DB, "<$FILE") or die "Can't open $FILE for read: $!\n";
    my $line;
    while (defined($line = <DB>)) {
	next if $line =~ /^\s*$/;
	$line =~ s/\s*#.*$//;
	chomp($line);
	my($sum, $file) = split(':', $line, 2);
	$db{$file} = $sum;
	print "Loading '$sum:$file' into database\n" if $DEBUG >= 2;
    }
    close(DB) or die "Can't close $FILE from read: $!\n";
    return %db;
}

# this sub takes an array reference full of files, and a hash reference
# with files pointing to sums, and then checks the actual sum of each file
# against the hash
sub test_files
{
    my($aref, $href) = @_;
    my $msg = '';
    my @files = @$aref;
    my %db = %$href;
    foreach my $file (@files) {
	print STDERR "Testing file '$file'" if $DEBUG >= 2;
	unless (exists($db{$file})) {
	    die "File ($file) not in database\n";
	}
	$file = &un_taint($file);
	my $sum = &checksum($file);
	unless ($sum eq $db{$file}) {
	    my $warning = "$file: didn't check out against the database.\n" .
		($DEBUG >= 1 ? "\tSum: '$sum'\n\t DB: '$db{$file}'\n" : "");
	    warn $warning;
	    $msg .= $warning;
	}
	print STDERR "\t\t(checked out)\n" if $DEBUG >= 2;
    }
    return $msg;
}

# this takes a possibly bad filename as a parameter and returns a sanitized
# version
sub un_taint
{
    my $file = shift;
    # require absolute pathnames
    unless ($file =~ /^\//) {
	croak "File '$file' is tainted: need absolute pathname, died";
    }
    # for now just allow alphanumerics, slashes dots, dashes, pluses, and
    # underscore
    if ($file =~ /^(\/[\w\/\.\-\+]+)$/) {
	$file = $1;
    }
    else {
	croak "File '$file' is tainted: only a-zA-Z0-9.-+_ allowed, died";
    }
    return $file;
}

# this takes the filename and list of files, and builds a database at filename
# with the check sum for each file
sub build_db
{
    my $db = shift;
    my @files = @_;
    if (-e $db) {
	# make sure it's not a symlink or any funkiness
	unless (-f $db and -w $db) {
	    die "Can't overwrite database at '$db',\n" .
		"not a regular file or do not have write permission.\n";
	}
	unlink($db);
    }
    open(DB, ">$db") or die "Can't create database '$db': $!\n";
    foreach my $file (@files) {
	$file = &un_taint($file);
	my $sum = &checksum($file);
	print DB "$sum:$file\n";
    }
    close(DB) or die "Can't close newly created database '$db': $!\n";
}

# send an email message to a userid
# Warning, both inputs to this function are assumed to be untainted
sub send_email
{
    (my $userid, my $msg) = @_;
    open(MAIL, "|$MAIL -s \"Mail from claymore\" \"$userid\"") or
	die "Can't open pipe to $MAIL: $!\n";
    print MAIL $msg;
    close(MAIL) or die "Can't close pipe to $MAIL: $!\n";
    return;
}

# This function performs the actual checksum on a specific file.
# It is passed a string, which is the path to the file, and returns a 32
# character hexadecimal checksum of the file's data. It uses the global
# Digest::MD5 object '$MD5' to avoid initializing a new object for every
# checksum.
sub checksum
{
    my $path = shift;
    &Digest::MD5::reset($MD5);

    # add the file contents
    open(DATA, $path) or die "Couldn't open '$path': $!";
    my $file = &Digest::MD5::addfile($MD5, *DATA);

    # add the permissions and uid/gid
    my @info = stat($path);
    $file = &Digest::MD5::add($MD5, $info[2]);
    $file = &Digest::MD5::add($MD5, $info[5]*65536 + $info[4]);

    my $sum = &Digest::MD5::hexdigest($file);
    #my $sum = $MD5->addfile(*DATA)->hexdigest;
    close(DATA) or die "Couldn't close '$path': $!";
    return $sum;
}
